import 'package:flutter/material.dart';
import 'dart:math' as math;

List<String> _prefixes = ['xl', 'lg', 'md', 'sm', ''];
List<String> _prefixesReversed = ['', 'sm', 'md', 'lg', 'xl'];

double _gutterSize = 10.0;
int _numberOfColumns = 24;
double _oneColumnRatio = 1.0 / _numberOfColumns;

void bootstrapGridParameters({
  int numberOfColumns = 24,
  double gutterSize = 24,
}) {
  assert(() {
    if (numberOfColumns < 10 || numberOfColumns > 24) {
      throw FlutterError.fromParts(<DiagnosticsNode>[
        ErrorSummary('Invalid number of columns: $numberOfColumns'),
        ErrorHint('The number of columns must be inside the [10; 24] range.'),
      ]);
    }

    if (gutterSize < 0.0 || gutterSize > 96.0) {
      throw FlutterError.fromParts(<DiagnosticsNode>[
        ErrorSummary('Invalid gutter size: $gutterSize'),
        ErrorHint('The gutterSize must be inside the [0.0; 96.0] range.'),
      ]);
    }

    return true;
  }());
  _numberOfColumns = numberOfColumns;
  _oneColumnRatio = 1.0 / numberOfColumns;
  _gutterSize = gutterSize;
}

String bootstrapPrefixBasedOnWidth(double width) {
  if (width > 1200) return "xl";

  if (width > 992) return "lg";

  if (width > 768) return "md";

  if (width > 575) return "sm";

  return '';
}

double bootstrapMaxWidthNonFluid(double width) {
  if (width >= 1200) return 1140;
  if (width >= 992) return 960;
  if (width >= 768) return 720;
  if (width >= 576) return 540;
  return width;
}

class BootstrapContainer extends StatelessWidget {
  BootstrapContainer({
    Key key,
    @required this.children,
    this.fluid = false,
    this.decoration,
    this.padding,
  }) : super(key: key);

  final bool fluid;
  final BoxDecoration decoration;
  final List<Widget> children;
  final EdgeInsets padding;
  double _getMaxWidth(double width) {
    if (this.fluid) return width;
    return bootstrapMaxWidthNonFluid(width);
  }

  @override
  Widget build(BuildContext context) {
    return LayoutBuilder(
      builder: (BuildContext context, BoxConstraints constraints) {
        double width = _getMaxWidth(constraints.maxWidth);

        Widget widget = Container(
          width: width,
          decoration: decoration,
          child: Wrap(
            alignment: WrapAlignment.start,
            direction: Axis.horizontal,
            children: children,
          ),
        );

        if (padding != null) {
          widget = Padding(
            padding: padding,
            child: widget,
          );
        }

        return Align(
          alignment: Alignment.topCenter,
          child: widget,
        );
      },
    );
  }
}

class BootstrapRow extends StatelessWidget {
  BootstrapRow({
    @required this.children,
    this.decoration,
    this.height,
  });
  final double height;
  final BoxDecoration decoration;
  final List<BootstrapCol> children;

  @override
  Widget build(BuildContext context) {
    return LayoutBuilder(
      builder: (BuildContext context, BoxConstraints constraints) {
        String pfx = bootstrapPrefixBasedOnWidth(constraints.maxWidth);
        List<BootstrapCol> _children = List.from(children);
        _children.sort((a, b) => a.orderPerSize[pfx] - b.orderPerSize[pfx]);

        return Container(
          width: constraints.maxWidth,
          height: height,
          decoration: decoration,
          child: Wrap(
            alignment: WrapAlignment.start,
            direction: Axis.horizontal,
            children: _children,
          ),
        );
      },
    );
  }
}

class BootstrapCol extends StatelessWidget {
  BootstrapCol({
    @required this.child,
    this.fit = FlexFit.loose,
    this.absoluteSizes = true,
    String sizes = "",
    String offsets = "",
    String orders = "",
    this.invisibleForSizes,
  })  : this.sizes = sizes.trim(),
        this.offsets = offsets.trim(),
        this.orders = orders.trim() {
    _initialize();
  }
  final Widget child;
  final FlexFit fit;

  ///
  /// Sizes
  ///
  /// Corresponds to the Bootstrap classes: .col-*
  ///
  /// Defines the dimensions (in terms of number of columns),
  /// depending on the screen width.
  ///
  /// Example:
  ///   sizes: 'col-12 col-sd-12 col-md-6 col-lg-4 col-xl-4'
  ///
  final String sizes;

  ///
  /// Offsets (= number of columns) to push this Widget to the right
  ///
  /// Corresponds to the Bootstrap classes: .offset-*
  ///
  /// Example:
  ///   offsets: 'offset-0 offset-sm-1 offset-md-2 offset-lg-3 offset-xl-4'
  ///
  final String offsets;

  ///
  /// Sequence orders
  ///
  /// Corresponds to the Bootstrap classes: .order-*
  ///
  /// Example:
  ///   orders: 'order-2 order-sm-2 order-md-1 order-lg-1 order-xl-1'
  ///
  final String orders;

  final String invisibleForSizes;
  final bool absoluteSizes;

  final Map<String, int> _ratiosPerSize = {
    'xl': 100,
    'lg': 100,
    'md': 100,
    'sm': 100,
    '': 100,
  };

  final Map<String, int> _offsetsPerSize = {
    'xl': -100,
    'lg': -100,
    'md': -100,
    'sm': -100,
    '': -100,
  };

  final Map<String, int> orderPerSize = {
    'xl': 0,
    'lg': 0,
    'md': 0,
    'sm': 0,
    '': 0,
  };

  final Map<String, bool> hiddenPerSize = {
    'xl': false,
    'lg': false,
    'md': false,
    'sm': false,
    '': false,
  };

  void _initialize() {
    final int nbPrefixes = _prefixes.length;

    void _initArray({
      String referenceArgument,
      Map<String, int> map,
      String argPrefix,
      Function minMaxFct,
      int lowerBoundValue = 0,
      int noValue,
      int minMaxNoValueReference,
    }) {
      List<String> parts = referenceArgument.isEmpty
          ? []
          : referenceArgument
              .toLowerCase()
              .split(' ')
              .where((t) => t.trim().isNotEmpty)
              .toList();
      parts.forEach((String part) {
        _prefixes.forEach((pfx) {
          final String prefix = '$argPrefix-$pfx${pfx == "" ? "" : "-"}';
          if (part.startsWith(prefix)) {
            String valString = part.split(prefix)?.last;
            if (valString != null && valString != prefix) {
              int value = int.tryParse(valString);
              if (value != null &&
                  value <= _numberOfColumns &&
                  value > lowerBoundValue) {
                map[pfx] = minMaxFct(map[pfx], value);
              }
            }
          }
        });
      });

      for (int idx = 0; idx < nbPrefixes; idx++) {
        String pfx = _prefixesReversed[idx];
        int value = map[pfx];

        if (value == noValue) {
          int i;
          for (i = idx + 1; i < nbPrefixes; i++) {
            String pfx2 = _prefixesReversed[i];
            if (map[pfx2] != noValue) {
              value = map[pfx2];
              break;
            }
          }

          if (value == noValue)
            for (int j = i - 1; j > -1; j--) {
              String pfx3 = _prefixesReversed[j];
              if (map[pfx3] != noValue) {
                value = map[pfx3];
                break;
              }
            }

          map[pfx] = minMaxFct(minMaxNoValueReference, value);
        }
      }
    }

    _initArray(
      referenceArgument: sizes,
      map: _ratiosPerSize,
      argPrefix: 'col',
      minMaxFct: math.min,
      noValue: 100,
      minMaxNoValueReference: _numberOfColumns,
    );

    _initArray(
      referenceArgument: offsets,
      map: _offsetsPerSize,
      argPrefix: 'offset',
      minMaxFct: math.max,
      noValue: -100,
      minMaxNoValueReference: 0,
      lowerBoundValue: -1,
    );

    _initArray(
      referenceArgument: orders,
      map: orderPerSize,
      argPrefix: 'order',
      minMaxFct: math.max,
      noValue: -100,
      minMaxNoValueReference: 0,
    );

    List<String> parts = (invisibleForSizes ?? "").trim().isEmpty
        ? []
        : invisibleForSizes
            .toLowerCase()
            .split(' ')
            .where((t) => t.trim().isNotEmpty)
            .toList();
    parts.forEach((String pfx) {
      if (['xl', 'lg', 'md', 'sm', 'xs'].contains(pfx)) {
        hiddenPerSize[pfx == 'xs' ? '' : pfx] = true;
      }
    });
  }

  int _getFlexRatio(String prefix) {
    return _ratiosPerSize[prefix];
  }

  int _getLeftMarginRatio(String prefix) {
    return _offsetsPerSize[prefix] ?? 0;
  }

  @override
  Widget build(BuildContext context) {
    return LayoutBuilder(
      builder: (BuildContext context, BoxConstraints constraints) {
        String pfx = bootstrapPrefixBasedOnWidth(absoluteSizes
            ? MediaQuery.of(context).size.width
            : constraints.maxWidth);
        bool isInvisible = hiddenPerSize[pfx];
        if (isInvisible) return Container();

        int flexRatio = _getFlexRatio(pfx);
        int leftMarginRatio = _getLeftMarginRatio(pfx);
        Widget widget = Container(
          width: flexRatio * constraints.maxWidth * _oneColumnRatio,
          child: Padding(
            padding: _gutterSize == 0.0
                ? EdgeInsets.zero
                : EdgeInsets.symmetric(horizontal: _gutterSize / 2),
            child: child,
          ),
        );

        if (leftMarginRatio > 0) {
          final double leftMargin =
              constraints.maxWidth * leftMarginRatio * _oneColumnRatio;
          widget = Padding(
            padding: EdgeInsets.only(left: leftMargin),
            child: widget,
          );
        }

        return widget;
      },
    );
  }
}

class BootstrapVisibility extends StatelessWidget {
  BootstrapVisibility({
    @required this.child,
    String sizes = "",
  }) : this.sizes = sizes.trim() {
    _initialize();
  }
  final Widget child;
  final String sizes;
  final Map<String, bool> _visibilityPerSize = {
    'xl': false,
    'lg': false,
    'md': false,
    'sm': false,
    '': false,
  };
  void _initialize() {
    List<String> parts = sizes.isEmpty
        ? []
        : sizes
            .toLowerCase()
            .split(' ')
            .where((t) => t.trim().isNotEmpty)
            .toList();
    parts.forEach((String part) {
      _prefixes.forEach((pfx) {
        final String prefix = 'col-$pfx';
        if (part.startsWith(prefix)) {
          _visibilityPerSize[pfx] = true;
        }
      });
    });
  }

  @override
  Widget build(BuildContext context) {
    String pfx = bootstrapPrefixBasedOnWidth(MediaQuery.of(context).size.width);
    if (pfx == '') {
      pfx = 'xs';
    }
    bool visible = _visibilityPerSize[pfx];

    if (visible == false) {
      return Container();
    }

    return child;
  }
}

dynamic bootStrapValueBasedOnSize({
  @required Map<String, dynamic> sizes,
  @required BuildContext context,
}) {
  String pfx = bootstrapPrefixBasedOnWidth(MediaQuery.of(context).size.width);

  final int nbPrefixes = _prefixes.length;
  dynamic value;
  value = sizes[pfx];

  if (value != null) return value;
  int i;
  int idx = _prefixes.indexOf(pfx);
  for (i = idx + 1; i < nbPrefixes; i++) {
    String pfx2 = _prefixesReversed[i];
    if (sizes[pfx2] != null) {
      value = sizes[pfx2];
      break;
    }
  }

  if (value != null) return value;
  for (int j = i - 1; j > -1; j--) {
    String pfx3 = _prefixesReversed[j];
    if (sizes[pfx3] != null) {
      value = sizes[pfx3];
      break;
    }
  }
  return value;
}
